#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

#include "bits.h"
#include "ftime.h"

/* Detect whether running on Alpha */
#ifdef __alpha
#define IS_ALPHA 1
#else
#define IS_ALPHA 0
#endif

/* Command Line options:
   -v N:  Set verbosity to level N
        0: Only give final scores
        1: Also report individual performance scores (default)
        2: Also report individual correctness scores
   -c: Run correctness tests only, not the performance tests
   -f Name: Check only the named function
   -e N: Limit number of errors to report for single function to N
         (Default unbounded)
*/

static int error_limit = 1000;

static char* test_fname = NULL;

/* Declare different function types */
typedef long int (*no_arg_fun) (void);
typedef long int (*single_fun)(long int);
typedef long int (*double_fun)(long int, long int); 

/* Dummy timing procedures */
long int dummy_no_arg(void);
long int dummy_single(long int);
long int dummy_double(long int, long int);


static volatile sink = 0;

/* Time-wasting code to slow down test_XXX */
#define SPIN_CNT 20
static void spin()
{
  int i;
  for (i = 0; i < SPIN_CNT; i++) {
    sink = i;
  }
}

/* Test versions of the functions */
long int test_bitOr(long int x, long int y)
{
  /* Waste some time */
  spin();
  return x|y;
}

long int test_bitXor(long int x, long int y)
{
  /* Waste some time */
  spin();
  return x^y;
}

long int test_isZero(long int x) {
  /* Waste some time */
  spin();
  return x == 0L;
}

long int test_isEqual(long int x, long int y)
{
  /* Waste some time */
  spin();
  return x == y; 
}

long int test_anyMaskedBits(long int x, long int y) 
{
  int i;
  long int result = 0L;
  for (i = 0; i < 64; i++) {
    if ((x & (1L << i)) && (y & (1L << i)))
      result = 1L;
  }
  return result;
}

long int test_allMaskedBits(long int x, long int y) 
{
  int i;
  long int result = 1L;
  for (i = 0; i < 64; i++) {
    if (!(x & (1L << i)) && (y & (1L << i)))
      result = 0L;
  }
  return result;
}

long int test_is32(void)
{
  /* Waste some time */
  spin();
  return 0xFFFFFFFFL == -1L;
}

long int test_signBit(void)
{
  /* Waste some time */
  spin();
  return test_is32() ? 31L : 63L;
}

long int test_highByte(long int x)
{
  long unsigned ux = x;
  long int shift = test_signBit() - 7L;
  unsigned long byte = ux >> shift;  
  /* Waste some time */
  spin();
  return (long int) byte;
}

long int test_TMin(void) {
  /* Waste some time */
  spin();
  return LONG_MIN;
}

long int test_minusOne(void) {
  /* Waste some time */
  spin();
  return -1L;
}

long int test_TMax(void) {
  /* Waste some time */
  spin();
  return LONG_MAX;
}

long int test_isNegative(long int x) {
  /* Waste some time */
  spin();
  return x < 0L;
}

long int test_isPositive(long int x) {
  /* Waste some time */
  spin();
  return x > 0L;
}

long int test_negate(long int x) {
  /* Waste some time */
  spin();
  return -x;
}

long int test_isGreater(long int x, long int y)
{
  /* Waste some time */
  spin();
  return x > y;
}

long int test_absval(long int x)
{
  /* Waste some time */
  spin();
  return (x < 0) ? -x : x; 
}

long int test_bang(long int x)
{
  /* Waste some time */
  spin();
  return !x;
}

long int dummy_no_arg(void)
{
  return 0L;
}

long int dummy_single(long int x)
{
  return 0L;
}

long int dummy_double(long int x, long int y)
{
  return 0L;
}


/* Assemble all of the information about different functions */
typedef struct {
	/* Function and test function */
	no_arg_fun f;
	no_arg_fun ft;
	/* Function name */
	char name[16];
	/* Maximum allowed clock cycles for full performance credit */
	int max_cycles;
	/* Measure of difficulty for this function */
	double difficulty;
} no_arg_info_rec, *no_arg_info_pointer;

typedef struct {
	/* Function and test function */
	single_fun f;
	single_fun ft;
	/* Function name */
	char name[16];
	/* Maximum allowed clock cycles for full performance credit */
	int max_cycles;
	/* Measure of difficulty for this function */
	double difficulty;
} single_arg_info_rec, *single_arg_info_pointer;

typedef struct {
	/* Function and test function */
	double_fun f;
	double_fun ft;
	/* Function name */
	char name[16];
	/* Maximum allowed clock cycles for full performance credit */
	int max_cycles;
	/* Measure of difficulty for this function */
	double difficulty;
} double_arg_info_rec, *double_arg_info_pointer;

#define NO_ARG_COUNT 5
no_arg_info_rec no_args[NO_ARG_COUNT] = {
  {is32, test_is32, "is32", 1,  2.0},
  {signBit, test_signBit, "signBit", 1, 3.0},
  {TMin, test_TMin, "TMin", 3,  1.0},
  {minusOne, test_minusOne, "minusOne", 1,  1.0},
  {TMax, test_TMax, "TMax", 3,  1.0}  
};

#define SINGLE_ARG_COUNT 7
single_arg_info_rec singles[SINGLE_ARG_COUNT] = {
  {isZero, test_isZero, "isZero", 2,  1.0},
  {highByte, test_highByte, "highByte", 2,  3.0},
  {isNegative, test_isNegative, "isNegative", 2,  2.0},
  {isPositive, test_isPositive, "isPositive", 3,  2.0},
  {negate, test_negate, "negate", 3,  2.0},
  {absval, test_absval, "absval", 3,  4.0},
  {bang, test_bang, "bang", 6,  4.0}
};
 
#define DOUBLE_ARG_COUNT 6
double_arg_info_rec doubles[DOUBLE_ARG_COUNT] = {
  {bitOr, test_bitOr, "bitOr", 1,  1.0},
  {bitXor, test_bitXor, "bitXor", 2,  2.0},
  {isEqual, test_isEqual, "isEqual", 2,  2.0},
  {anyMaskedBits, test_anyMaskedBits, "anyMaskedBits", 3,  2.0},
  {allMaskedBits, test_allMaskedBits, "allMaskedBits", 3,  2.0},
  {isGreater, test_isGreater, "isGreater", 9,  3.0}
};

/* Generate test values near "corner cases" */
#define TEST_RANGE 4
#define TEST_COUNT (6*TEST_RANGE)

long int test_vals[TEST_COUNT];

static void gen_vals()
{
  int i;
  for (i = 0; i < TEST_RANGE; i++) {
    test_vals[i] = LONG_MIN+i;
    test_vals[i+TEST_RANGE] = SHRT_MIN-i;
    test_vals[i+2*TEST_RANGE] = -i;
    test_vals[i+3*TEST_RANGE] = i;
    test_vals[i+4*TEST_RANGE] = SHRT_MAX+i;
    test_vals[i+5*TEST_RANGE] = LONG_MAX-i;
  }
}

/* Test no argument function f against know good ft */
static int test_no_arg(no_arg_fun f, no_arg_fun ft,
		       char *description, int report)
{
  int errors = 0;
  long int r = f();
  long int rt = ft();
  if (r != rt) {
    errors++;
    if (report && errors <= error_limit) { 
      printf("Test %s() failed.  Gives %ld.  Should be %ld\n",
	       description, r, rt);
    }
  }
  return errors;
}

/* Test single argument function f against known good ft */
static int test_single(single_fun f, single_fun ft,
		       char *description, int report)
{
  int i;
  int errors = 0;
  for (i = 0; i < TEST_COUNT; i++) {
    long int x = test_vals[i];
    long int r = f(x);
    long int rt = ft(x);
    if (r != rt) {
      errors++;
      if (report && errors <= error_limit)  
        printf("Test %s(%ld) failed.  Gives %ld.  Should be %ld\n",
	       description, x, r, rt);
    }
  }
  return errors;
}

/* Test double argument function f against known good ft */
static int test_double(double_fun f, double_fun ft,
		       char *description, int report)
{
  int i, j;
  int errors = 0;
  for (i = 0; i < TEST_COUNT; i++) {
    for (j = 0; j < TEST_COUNT; j++) {
      long int x = test_vals[i];
      long int y = test_vals[j];
      long int r = f(x,y);
      long int rt = ft(x,y);
      if (r != rt) {
        errors++;
        if (report && errors <= error_limit)
          printf("Test %s(%ld,%ld) failed.  Gives %ld.  Should be %ld\n",
	         description, x, y, r, rt);
	  }
    }
  }
  return errors;
}

/* Timer Stuff */
static double clock_freq = 0.0;

static no_arg_fun f0 = NULL;
static single_fun f1 = NULL;
static double_fun f2 = NULL;

/* Compute overheads of different functions */
static double f0_over = 0.0;
static double f1_over = 0.0;
static double f2_over = 0.0;



/* How many times does each function get called? */
#define REPEAT_COUNT 5

/* Run no argument function */
static void run_no_arg()
{
  int i, j;
  for (i = 0; i < TEST_COUNT; i++)
    for (j = 0; j < TEST_COUNT; j++) {
#if 0
      long int x = test_vals[i];
      long int y = test_vals[j];
#endif
      /* Run total of 5 times */
      long int r = 0L;
      r += f0();
      r += f0();
      r += f0();
      r += f0();
      r += f0();
      sink = r; 
    }
}

/* Run single argument function */
static void run_single()
{
  int i, j;
  for (i = 0; i < TEST_COUNT; i++)
    for (j = 0; j < TEST_COUNT; j++) {
      long int x = test_vals[i];
#if 0
      long int y = test_vals[j];
#endif
      /* Run total of 5 times */
      long int r = 0L;
      r += f1(x);
      r += f1(x);
      r += f1(x);
      r += f1(x);
      r += f1(x);
      sink = r; 
    }
}

/* Run two argument function */
static void run_double()
{
  int i, j;
  for (i = 0; i < TEST_COUNT; i++)
    for (j = 0; j < TEST_COUNT; j++) {
      long int x = test_vals[i];
      long int y = test_vals[j];
      /* Run total of 5 times */
      long int r = 0L;
      r += f2(x,y);
      r += f2(x,y);
      r += f2(x,y);
      r += f2(x,y);
      r += f2(x,y);
      sink = r; 
    }
}

#define ETOL 0.02

/* Initialize timer */
static void init_timer()
{
  if (clock_freq == 0.0) {
    clock_freq = freq();
    /* Compute overheads */
    f0 = dummy_no_arg;
    f0_over = ftime(run_no_arg, ETOL);
    f1 = dummy_single;
    f1_over = ftime(run_single, ETOL);
    f2 = dummy_double;
    f2_over = ftime(run_double, ETOL);
  }
}


/* Time no argument function f */
static double time_no_arg(no_arg_fun f, char *description, int report)
{
  double time;
  double cycles;
  init_timer();
  f0 = f;
  time = (ftime(run_no_arg, ETOL) - f0_over) /
    (TEST_COUNT * TEST_COUNT * REPEAT_COUNT) * 1e6;
  if (time < 0.0)
    time = 0.0;
  cycles = time * clock_freq;
#ifdef GRADE
  printf("\t%.2f", cycles);
#else  
  if (report)
    printf("Test %s:\t%.2f cycles", description, cycles);
#endif
  return cycles;
}

/* Time single argument function f */
static double time_single(single_fun f, char *description, int report)
{
  double time;
  double cycles;
  init_timer();
  f1 = f;
  time = (ftime(run_single, ETOL) - f1_over) /
    (TEST_COUNT * TEST_COUNT * REPEAT_COUNT) * 1e6;
  if (time < 0.0)
    time = 0.0;
  cycles = time * clock_freq;
#ifdef GRADE
  printf("\t%.2f", cycles);
#else  
  if (report)
    printf("Test %s:\t%.2f cycles", description, cycles);
#endif
  return cycles;
}

/* Time two argument function f */
static double time_double(double_fun f, char *description, int report)
{
  double time;
  double cycles;
  init_timer();
  f2 = f;
  time = (ftime(run_double, ETOL) - f2_over) /
    (TEST_COUNT * TEST_COUNT * REPEAT_COUNT) * 1e6;
  if (time < 0.0)
    time = 0.0;
  cycles = time * clock_freq;
#ifdef GRADE
  printf("\t%.2f", cycles);
#else  
  if (report)
    printf("Test %s:\t%.2f cycles", description, cycles);
#endif
  return cycles;
}


/* Run series of tests.  Return number of errors */ 
static int run_tests(int report) {
  int t;
  int errors = 0;
  double points = 0.0;
  double max_points = 0.0;
  gen_vals();
  for (t = 0; t < NO_ARG_COUNT; t++) {
    int terrors;
    double tscore;
    double tpoints;
  	no_arg_fun f = no_args[t].f;
  	no_arg_fun ft = no_args[t].ft;
  	char *fname = no_args[t].name;
  	if (!test_fname || strcmp(fname, test_fname) == 0) {
      terrors = test_no_arg(f, ft, fname, 1);
#ifdef GRADE
      printf("\t%d", terrors);
#endif
      errors += terrors;
      tscore = terrors ? 0.0 : 1.0;
      tpoints = no_args[t].difficulty * tscore;
      points += tpoints;
      max_points += no_args[t].difficulty;
      if (report)
        printf("Test %s score: %.2f/%.2f\n",
	       fname, tpoints, no_args[t].difficulty);
    }
  }
  for (t = 0; t < SINGLE_ARG_COUNT; t++) {
    int terrors;
    double tscore;
    double tpoints;
  	single_fun f = singles[t].f;
  	single_fun ft = singles[t].ft;
  	char *fname = singles[t].name;
  	if (!test_fname || strcmp(fname, test_fname) == 0) {
      terrors = test_single(f, ft, fname, 1);
#ifdef GRADE
      printf("\t%d", terrors);
#endif
      errors += terrors;
      tscore = terrors == 0 ? 1.0 : terrors <= 1 ? 0.5 : 0.0;
      tpoints = singles[t].difficulty * tscore;
      points += tpoints;
      max_points += singles[t].difficulty;
      if (report)
        printf("Test %s score: %.2f/%.2f\n",
	       fname, tpoints, singles[t].difficulty);
    }
  }
  for (t = 0; t < DOUBLE_ARG_COUNT; t++) {
    int terrors;
    double tscore;
    double tpoints;
  	double_fun f = doubles[t].f;
  	double_fun ft = doubles[t].ft;
  	char *fname = doubles[t].name;
  	if (!test_fname || strcmp(fname, test_fname) == 0) {
      terrors = test_double(f, ft, fname, 1);
#ifdef GRADE
      printf("\t%d", terrors);
#endif
      errors += terrors;
      tscore = terrors == 0 ? 1.0 : terrors <= 1 ? 0.5 : 0.0;
      tpoints = doubles[t].difficulty * tscore;
      points += tpoints;
      max_points += doubles[t].difficulty;
      if (report)
        printf("Test %s score: %.2f/%.2f\n",
	       fname, tpoints, doubles[t].difficulty);
    }
  }
#ifdef GRADE
  printf("\t%.2f", points);
#else
  printf("Overall Correctness score: %.2f/%.2f\n", points, max_points);
#endif
  return errors;
}

/* Compute performance score for problem */
static double perf_points(double cycles, double limit, 
  double difficulty, int report)
{
    double tscore, tpoints;
	if (cycles - 0.005 < limit)
	  tscore = 1.0;
    else if (cycles - 0.005 < 1.5 * limit)
      tscore = 2.0/3.0;
    else if (cycles - 0.005 < 2.0 * limit)
      tscore = 1.0/3.0;
    else
      tscore = 0.0;
    tpoints = difficulty * tscore;
    if (report)
      printf(" (%.1f goal)\tScore: %.2f/%.2f\n", limit, tpoints, difficulty);
    return tpoints;
}

/* Time functions */
static void run_times(int report) {
  int t;
  double points = 0.0;
  double max_points = 0.0;
  gen_vals();
  for (t = 0; t < NO_ARG_COUNT; t++) {
    int terrors;
    double tpoints =  0.0;
    double difficulty = no_args[t].difficulty/2.0;
  	no_arg_fun f = no_args[t].f;
  	no_arg_fun ft = no_args[t].ft;
  	char *fname = no_args[t].name;
  	if (!test_fname || strcmp(test_fname, fname) == 0) {
      terrors = test_no_arg(f, ft, fname, 0);
      if (terrors == 0) {
        double cycles = time_no_arg(f, fname, report);
        double limit = no_args[t].max_cycles;
        tpoints = perf_points(cycles, limit, difficulty, report);
      } else {
#ifdef GRADE
	printf("\tNA");
#endif
      }
      points += tpoints;
      max_points += difficulty;
    }
  }
  for (t = 0; t < SINGLE_ARG_COUNT; t++) {
    int terrors;
    double tpoints =  0.0;
    double difficulty = singles[t].difficulty/2.0;
  	single_fun f = singles[t].f;
  	single_fun ft = singles[t].ft;
  	char *fname = singles[t].name;
  	if (!test_fname || strcmp(test_fname, fname) == 0) {
      terrors = test_single(f, ft, fname, 0);
      if (terrors == 0) {
        double cycles = time_single(f, fname, report);
        double limit = singles[t].max_cycles;
        tpoints = perf_points(cycles, limit, difficulty, report);
      } else {
#ifdef GRADE
	printf("\tNA");
#endif
      }
      points += tpoints;
      max_points += difficulty;
    }
  }
  for (t = 0; t < DOUBLE_ARG_COUNT; t++) {
    int terrors;
    double tpoints =  0.0;
    double difficulty = doubles[t].difficulty/2.0;
  	double_fun f = doubles[t].f;
  	double_fun ft = doubles[t].ft;
  	char *fname = doubles[t].name;
  	if (!test_fname || strcmp(test_fname, fname) == 0) {
      terrors = test_double(f, ft, fname, 0);
      if (terrors == 0) {
        double cycles = time_double(f, fname, report);
        double limit = doubles[t].max_cycles;
        tpoints = perf_points(cycles, limit, difficulty, report);
      } else {
#ifdef GRADE
	printf("\tNA");
#endif
      }
      points += tpoints;
      max_points += difficulty;
    }
  }
#ifdef GRADE
  printf("\t%.2f", points);
#else
  printf("Overall Performance score: %.2f/%.2f\n", points, max_points);
#endif
}

/* Command Line options:
   -v N:  Set verbosity to level N
      0: Only give final scores
      1: Also report individual performance scores (default)
      2: Also report individual correctness scores
   -c: Run correctness tests only, not the performance tests
   -f Name: Check only the named function
*/

static char *cmd = "btest";
   
static void usage() {
  printf("%s [-v 0|1|2] [-c] [-f Name] [-e ELIM]\n", cmd);
  exit(1);
}

extern team_struct team;

int main(int argc, char *argv[])
{
  int i;
  int do_perf = IS_ALPHA;
  int verbose_level = 1;
  int errors;
  cmd = argv[0];
  
#ifdef GRADE
  verbose_level = 0;
  error_limit = 0;
#endif  

  for (i = 1; i < argc; i++) {
    char* arg = argv[i];
    if (arg[0] == '-') {
      switch(arg[1]) {
        char cval;
      case 'c':
	do_perf = 0;
	break;
      case 'v':
	cval = arg[2];
	if (cval == '\0') {
	  if (i < argc-1) {
	    i++;
	    arg = argv[i];
	    cval = arg[0];
	  } else
	    usage();
	}
	if (cval == '0')
	  verbose_level = 0;
	else if (cval == '1')
	  verbose_level = 1;
	else if (cval == '2') 
	  verbose_level = 2;
	else
	  usage();
	break;
      case 'f':
	test_fname = arg+2;
	if (*test_fname == '\0') {
	  if (i < argc-1) {
	    i++;
	    test_fname = argv[i];
	  } else
	    usage();
	}
	break;
      case 'e':
	/* Set error limit */
	arg+=2;
	if (*arg == '\0') {
	  if (i < argc-1) {
	    i++;
	    arg = argv[i];
	  } else
	    usage();
	}
	error_limit = atoi(arg);
	if (error_limit < 0)
	  usage();
	break;
      default:
	usage();
      }
    } else
      usage();
  }
#ifdef GRADE
  printf("%sAlpha", IS_ALPHA ? "" : "not"); 
#else
  printf("Running btest on %s\n", IS_ALPHA ? "Alpha" : "Non-Alpha");
#endif
  if(*team.name1 == '\0') {
    printf("ERROR.  Must fill in team member names!\n");
    exit(0);
  } else {
#ifdef GRADE
    printf("\t%s", team.id1);
#else
    printf("Member 1:\t%s\t%s\n", team.name1, team.id1);
#endif
  }
  if (*team.name2 != '\0') {
#ifdef GRADE
    printf("\t%s", team.id2);
#else
    printf("Member 2:\t%s\t%s\n", team.name2, team.id2);
#endif
  } else {
#ifdef GRADE
    printf("\t\"\"");
#endif
  }
  errors = run_tests(verbose_level > 1);
#ifndef GRADE
  if (errors > 0)
    printf("%d errors encoufntered.\n", errors);
  else {
    printf("All tests passed.\n");
  }
#endif
  if (do_perf)
    run_times(verbose_level > 0);
#ifdef GRADE
  printf("\n");
#endif
  return 0;
}
     
